"""
Abstract Interface for scenario models.

Classes
-------
- ScenarioModel: Abstract base class for scenario models.
"""

from abc import abstractmethod
from typing import Optional, Tuple, ClassVar, Any
from typing_extensions import Self  # python 3.9 & 3.10 compatibility

from pydantic import (
    Field,
    computed_field,
    model_validator,
    field_validator,
    AliasChoices,
    PrivateAttr,
    FieldSerializationInfo,
    field_serializer,
    SerializerFunctionWrapHandler,
    model_serializer,
    SerializationInfo,
)
from numpydantic import NDArray, Shape

import numpy as np
from pyRadPlan.core import PyRadPlanBaseModel
from pyRadPlan.ct import validate_ct


class ScenarioModel(PyRadPlanBaseModel):
    """Abstract base class for scenario models.

    Attributes
    ----------
    range_rel_sd : float
        The relative standard deviation for range.
    range_abs_sd : float
        The absolute standard deviation for range.
    shift_sd : Tuple[float]
        The standard deviation for shift in x, y, and z directions.
    wc_sigma : float
        The sigma value for the wc_factor.
    ct_scen_prob : list[Tuple[int,float]]
        The probability of each CT scenario.

    Methods
    -------
        update_scenarios() -> np.ndarray[float]
            Abstract method to update scenarios.
        extract_single_scenario(scen_num: int) -> ScenarioModel
            Extracts a single scenario based on its number.
        list_all_scenarios()
            Lists all scenarios.
        sub2scenIx(ctScen: int, shift_scen: int, range_shift_scen: int) -> int
            Converts CT scenario, shift scenario, and range shift scenario to scenario index.
        scenNum(fullScenIx: int) -> int
            Converts full scenario index to scenario number.
    """

    name: ClassVar[str]
    short_name: ClassVar[str]

    range_rel_sd: float = Field(
        default=3.5,
        description="The relative standard deviation for range.",
        ge=0,
        alias="rangeRelSD",
    )
    range_abs_sd: float = Field(
        default=1,
        description="The absolute standard deviation for range.",
        ge=0,
        alias="rangeAbsSD",
    )
    shift_sd: Tuple[float, float, float] = Field(
        default=(2.25, 2.25, 2.25),
        description="The standard deviation for shift in x, y, and z directions.",
        alias="shiftSD",
    )
    ct_scen_prob: list[Tuple[int, float]] = Field(
        default=[(0, 1.0)], description="The probability of each CT scenario."
    )
    wc_sigma: float = Field(
        1,
        description="Worst-case definition in multiples of standard deviation.",
        ge=0,
        validation_alias=AliasChoices("wcSigma", "wc_factor", "wcFactor"),
        serialization_alias="wcSigma",
    )

    # Other properties with protected access
    _num_of_available_ct_scen: int = PrivateAttr(default=1)
    _ct_scen_ix: np.ndarray[int] = PrivateAttr(default=np.array([0]))
    _iso_shift: np.ndarray[float] = PrivateAttr(default=np.array([[0.0, 0.0, 0.0]]))
    _rel_range_shift: np.ndarray[float] = PrivateAttr(default=0.0)
    _abs_range_shift: np.ndarray[float] = PrivateAttr(default=0.0)
    _tot_num_shift_scen: int = PrivateAttr(default=1)
    _tot_num_range_scen: int = PrivateAttr(default=1)
    _tot_num_scen: int = PrivateAttr(default=1)
    _scen_for_prob: np.ndarray[float] = PrivateAttr(
        default=np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])
    )
    _scen_prob: np.ndarray[float] = PrivateAttr(default=np.array([1.0]))
    _scen_weight: np.ndarray[float] = PrivateAttr(default=np.array([1.0]))
    _scen_mask: np.ndarray[bool] = PrivateAttr(default=np.ones((1, 1, 1), dtype=bool))
    _linear_mask: np.ndarray[int] = PrivateAttr(default=np.array([[0, 0, 0]]))

    # Constructor
    def __init__(self, ct: Optional[Any] = None, **kwargs):
        super(ScenarioModel, self).__init__(**kwargs)

        if ct is not None:
            ct = validate_ct(ct)
            self._num_of_available_ct_scen = ct.num_of_ct_scen
        else:
            self._num_of_available_ct_scen = 1

    @model_validator(mode="after")
    def _validate_model(self) -> Self:
        """
        Post-assignment validator.

        Mainly used to update Scenario Information.
        """
        self.update_scenarios()
        return self

    @field_validator("ct_scen_prob", mode="after")
    @classmethod
    def _validate_ct_scen_prob(
        cls, ct_scen_prob: list[Tuple[int, float]]
    ) -> list[Tuple[int, float]]:
        """
        Post-assignment validator for ct_scen_prob.

        Ensures sensible probabilities for ct scnearios.
        """
        if not all(0 <= x[1] <= 1 for x in ct_scen_prob):
            raise ValueError("All probabilities must be between 0 and 1.")
        if not all(0 <= x[0] for x in ct_scen_prob):
            raise ValueError("Scenario indices must be positive.")

        return ct_scen_prob

    @field_serializer("ct_scen_prob")
    def _serialize_ct_scen_prob(
        self, val: Any, info: FieldSerializationInfo
    ) -> list[Tuple[int, float]]:
        """Adapt for matRad if serialized for matRad."""

        if info.context is not None and "matRad" in info.context:
            return [(a + 1, b) for a, b in val]

        return val

    @computed_field
    @property
    def num_of_ct_scen(self) -> int:
        """Number of CT scenarios in the model."""
        return len(self.ct_scen_prob)

    @computed_field
    @property
    def num_of_available_ct_scen(self) -> int:
        """Number of totally available CT scenarios in the CT."""
        return self._num_of_available_ct_scen

    @computed_field
    @property
    def ct_scen_ix(self) -> NDArray[Shape["1-*"], int]:
        """CT scenario indices of the used CT scenarios."""
        return self._ct_scen_ix

    @computed_field
    @property
    def iso_shift(self) -> NDArray[Shape["1-*,3"], float]:
        """Isocenter shift values."""
        return self._iso_shift

    @computed_field
    @property
    def rel_range_shift(self) -> float:
        """Relative range shift value."""
        return self._rel_range_shift

    @computed_field
    @property
    def abs_range_shift(self) -> float:
        """Absolute range shift value."""
        return self._abs_range_shift

    @computed_field
    @property
    def max_abs_range_shift(self) -> float:
        """Maximum absolute range shift value."""
        return np.max(self.abs_range_shift)

    @computed_field
    @property
    def max_rel_range_shift(self) -> float:
        """Maximum relative range shift value."""
        return np.max(self.abs_range_shift)

    @computed_field
    @property
    def tot_num_shift_scen(self) -> int:
        """Total number of shift scenarios."""
        return self._tot_num_shift_scen

    @computed_field
    @property
    def tot_num_range_scen(self) -> int:
        """Total number of range shift scenarios."""
        return self._tot_num_range_scen

    @computed_field
    @property
    def tot_num_scen(self) -> int:
        """Total number of scenarios."""
        return self._tot_num_scen

    @computed_field
    @property
    def scen_for_prob(self) -> NDArray[Shape["1-*,1-*"], float]:
        """Scenarios organized for probability calculation."""
        return self._scen_for_prob

    @computed_field
    @property
    def scen_prob(self) -> NDArray[Shape["1-*,1-*"], float]:
        """Scenarios probability matrix."""
        return self._scen_prob

    @computed_field
    @property
    def scen_weight(self) -> NDArray[Shape["1-*,1-*"], float]:
        """Scenarios weight matrix."""
        return self._scen_weight

    @computed_field
    @property
    def scen_mask(self) -> NDArray[Shape["1-*,1-*"], bool]:
        """Scenarios mask matrix describing how scenarios should be stored."""
        return self._scen_mask

    @computed_field
    @property
    def linear_mask(self) -> NDArray[Shape["1-*,1-*"], int]:
        """Linear mask matrix for scenario selection."""
        return self._linear_mask

    @field_serializer("linear_mask")
    def _serialize_linear_mask(
        self, val: Any, info: FieldSerializationInfo
    ) -> NDArray[Shape["1-*,1-*"], int]:
        """Serialize the linear mask according to context."""

        if info.context is not None and "matRad" in info.context:
            return val + 1

        return val

    # Abstract methods
    @abstractmethod
    def update_scenarios(self) -> np.ndarray[float]:
        """Update scenario data from uncertainty model settings."""

    @abstractmethod
    def extract_single_scenario(self, scen_num: int) -> "ScenarioModel":
        """Extract a single scenario based on its number."""

    def list_all_scenarios(self):
        """
        Print a list of all scenarios.

        This method prints the details of all scenarios in a formatted table.
        The table includes the following columns:
        - xShift: The x-axis shift value
        - yShift: The y-axis shift value
        - zShift: The z-axis shift value
        - absRng: The absolute range value
        - relRng: The relative range value
        - prob: The probability value

        Example output:
        #   xShift   yShift   zShift   absRng   relRng   prob.
        1   0.000    0.000    0.000    0.000    0.000    0.000
        2   0.100    0.200    0.300    0.400    0.500    0.600
        3   0.700    0.800    0.900    1.000    1.100    1.200
        """

        print("Listing all scenarios...")
        print("#\tctScen\txShift\tyShift\tzShift\tabsRng\trelRng\tprob.")
        for s, row in enumerate(self.scen_for_prob):
            row_str = "\t".join([f"{int(row[0]):5d}"] + [f"{x:.3f}" for x in row[1:]])
            print(f"{s + 1}\t{row_str}\t{self.scen_prob[s]:.3f}")

    def sub2scen_ix(self, ct_scen: int, shift_scen: int, range_shift_scen: int) -> int:
        """
        Convert a subscript index to a linear scenario index.

        Parameters
        ----------
            ctScen (int): The sub-scenario index.
            shift_scen (int): The shift scenario value.
            range_shift_scen (int): The range shift scenario value.

        Returns
        -------
            int: The corresponding scenario index.

        Raises
        ------
            None

        Examples
        --------
            # Example usage
            sub2scenIx(1, 2, 3)  # Returns: 1
        """

        if self._scen_mask.shape[0] == 1:
            return self._ct_scen_ix[ct_scen]
        linear_index = np.ravel_multi_index(
            (ct_scen, shift_scen, range_shift_scen), self._scen_mask.shape
        )
        return linear_index

    def scen_num(self, full_scen_ix: int) -> int:
        """
        Return the scenario number given the scecnario ray index.

        This corresponds to the first occurrence of `full_scen_ix` in
        `self.scen_mask`.

        Parameters
        ----------
        full_scen_ix : int
            The value to search for in `self.scen_mask`.

        Returns
        -------
        int
            The index of the first occurrence of `full_scen_ix` in `self.scen_mask`.
        """

        scen_indices = np.where(self._scen_mask.flatten())[0]
        return int(np.where(scen_indices == full_scen_ix)[0][0])

    @model_serializer(mode="wrap")
    def _serialize(self, wrap: SerializerFunctionWrapHandler, info: SerializationInfo) -> dict:
        """
        Serialize the model to a dictionary.

        Parameters
        ----------
        wrap : SerializerFunctionWrapHandler
            The serialization function wrapper.

        Returns
        -------
        dict
            The serialized model.
        """

        output = wrap(self, info)
        output["model"] = self.short_name
        return output



import pytest

from pyRadPlan.dose.engines import (
    ParticleHongPencilBeamEngine,
    get_available_engines,
    get_engine,
)


def test_available_engines(test_data_protons):
    engines = get_available_engines(test_data_protons[0])
    assert isinstance(engines, dict)
    assert len(engines.items()) > 0


def test_get_engine(test_data_protons):
    engine = get_engine(test_data_protons[0])
    assert engine
    assert isinstance(engine, ParticleHongPencilBeamEngine)


def test_get_engine_default(test_data_protons):
    engine = get_engine(test_data_protons[0])
    assert engine
    assert isinstance(engine, ParticleHongPencilBeamEngine)


def test_get_engine_invalid(test_data_protons):
    test_data_protons[0].prop_dose_calc["engine"] = "InvalidEngine"
    with pytest.raises(ValueError):
        engine = get_engine(test_data_protons[0])



from typing import Any, Optional, Annotated, Union, Final
from typing_extensions import Self

# import warnings
import numpy as np
from pydantic import (
    Field,
    StringConstraints,
    model_validator,
    field_validator,
    # ValidatorFunctionWrapHandler,
    ValidationInfo,
    ValidationError,
    AliasChoices,
    computed_field,
)
from numpydantic import NDArray, Shape
from pyRadPlan.core import PyRadPlanBaseModel
from pyRadPlan.util.helpers import dl2ld
from pyRadPlan.machines._base import ExternalBeamMachine


class IonBeamEmittance(PyRadPlanBaseModel):
    """Emittance data for ion accelerator."""

    type: Final[str] = "bigaussian"

    sigma_x: float = Field(alias="sigmaX", ge=0.0)
    sigma_y: float = Field(alias="sigmaY", ge=0.0)
    div_x: float = Field(default=1e-12, alias="divX")
    div_y: float = Field(default=1e-12, alias="divY")
    corr_x: float = Field(default=1e-12, alias="corrX", ge=-1.0, le=1.0)
    corr_y: float = Field(default=1e-12, alias="corrY", ge=-1.0, le=1.0)


class IonEnergySpectrum(PyRadPlanBaseModel):
    pass


class IonEnergySpectrumGaussian(IonEnergySpectrum):
    """
    Energy spectrum data for ion pencil beam kernels.

    Notes
    -----
    The energy spectrum is defined as a Gaussian distribution with a mean
    energy and a relative spread in percent.
    """

    type: Final[str] = "gaussian"

    mean: float = Field(ge=0.0)
    sigma: float = Field(
        ge=0.0, description="Relative spread as percentage sigma of the Gaussian energy spectrum"
    )

    @property
    def sigma_relative(self) -> float:
        """Relative spread of the energy spectrum."""
        return self.sigma / 100.0

    @property
    def sigma_absolute(self) -> float:
        """Absolute sigma of the energy spectrum."""
        return self.sigma_relative * self.mean

    @property
    def fwhm(self) -> float:
        """Full width at half maximum of the energy spectrum."""
        return 2.35482 * self.sigma_absolute


class IonBeamFocus(PyRadPlanBaseModel):
    """Focus data for ion pencil beam kernels."""

    dist: NDArray[Shape["1-*"], np.float64]
    sigma: NDArray[Shape["1-*"], np.float64]
    fwhm_iso: Optional[np.float64] = None

    # emittance parameterization
    emittance: Optional[IonBeamEmittance] = None

    @field_validator("dist", "sigma", mode="before")
    @classmethod
    def validate_arrays(cls, v: Any) -> Any:
        """Validate the focus distance and sigma arrays."""
        try:
            v = np.array(v, dtype=np.float64)
        except ValueError as exc:
            raise exc

        return v

    @property
    def has_emittance(self) -> bool:
        """Check if emittance parameters are available."""
        return self.sigma_x is not None and self.sigma_y is not None


class LateralCutOff(PyRadPlanBaseModel):
    """Lateral cut-off radius and compensation for ion pencil beam kernels."""

    comp_fac: np.float64 = Field(default=np.float64(1.0))
    cut_off: NDArray[Shape["1-*"], np.float64] = Field(
        default=np.array([np.inf, np.inf], dtype=np.float64)
    )
    depths: NDArray[Shape["1-*"], np.float64] = Field(
        default=np.array([0.0, np.inf], dtype=np.float64)
    )

    @field_validator("comp_fac", mode="before")
    @classmethod
    def validate_comp_fac(cls, v: Any) -> Any:
        """Validate the lateral cut-off compensation factor."""
        try:
            return np.float64(v)
        except ValueError as exc:
            raise exc

    @field_validator("cut_off", mode="before")
    @classmethod
    def validate_arrays(cls, v: Any) -> Any:
        """Validate the lateral cut-off radial distance and depths."""
        try:
            v = np.array(v, dtype=np.float64)
        except ValueError as exc:
            raise exc

        return v


class IonPencilBeamKernel(PyRadPlanBaseModel):
    """Data Model for a single Ion Pencil Beam Kernel."""

    energy: np.float64
    peak_pos: np.float64
    range: Optional[np.float64] = None
    offset: np.float64 = Field(default=np.float64(0.0))
    depths: NDArray[Shape["1-*"], np.float64]

    # all kernels
    idd: NDArray[Shape["1-*"], np.float64] = Field(
        validation_alias=AliasChoices("Z", "z"), serialization_alias="Z"
    )
    sigma: Optional[NDArray[Shape["1-*"], np.float64]] = None
    sigma_1: Optional[NDArray[Shape["1-*"], np.float64]] = None
    sigma_2: Optional[NDArray[Shape["1-*"], np.float64]] = None
    weight: Optional[NDArray[Shape["1-*"], np.float64]] = None
    sigma_multi: Optional[NDArray[Shape["1-*,1-*"], np.float64]] = None
    weight_multi: Optional[NDArray[Shape["1-*,1-*"], np.float64]] = None
    let: Optional[NDArray[Shape["1-*"], np.float64]] = Field(alias="LET", default=None)
    alpha_x: Optional[NDArray[Shape["1-*"], np.float64]] = None
    beta_x: Optional[NDArray[Shape["1-*"], np.float64]] = None
    alpha: Optional[NDArray[Shape["1-*,1-*"], np.float64]] = None
    beta: Optional[NDArray[Shape["1-*,1-*"], np.float64]] = None
    lateral_cut_off: Optional[LateralCutOff] = None

    @field_validator("energy", "peak_pos", "range", "offset", mode="before")
    @classmethod
    def validate_possible_cast(cls, v: Any) -> Any:
        """Validate if the input can be cast to a float64."""
        try:
            return np.float64(v)
        except ValueError:
            return v

    @field_validator(
        "depths",
        "idd",
        "sigma",
        "sigma_1",
        "sigma_2",
        "weight",
        "sigma_multi",
        "weight_multi",
        "let",
        "alpha",
        "beta",
        "alpha_x",
        "beta_x",
        mode="before",
    )
    @classmethod
    def validate_possible_cast_array(cls, v: Any) -> Any:
        """Validate if the input can be cast to a float64 array."""
        try:
            v = np.array(v, dtype=np.float64)
        except ValueError as exc:
            raise exc

        return v

    @field_validator("idd", "sigma", "sigma_1", "sigma_2", "let", mode="after")
    @classmethod
    def validate_kernel_lengths(
        cls, v: Union[np.ndarray, None], info: ValidationInfo
    ) -> Union[np.ndarray, None]:
        """Validate the length of the kernel data."""
        if v is None:
            return v

        if v.shape != info.data["depths"].shape:
            raise ValueError("Kernel data length does not match the depth data length.")

        return v

    @field_validator("sigma_multi", "weight_multi", mode="after")
    @classmethod
    def validate_multi_gaussian_kernel(
        cls, v: Union[np.ndarray, None], info: ValidationInfo
    ) -> Union[np.ndarray, None]:
        """Validate the length of the multi-gaussian kernel data."""
        if v is None:
            return v

        if v.shape[0] != info.data["depths"].shape[0]:
            raise ValueError("Kernel data length does not match the depth data length.")

        return v

    @field_validator("alpha", "beta", mode="after")
    @classmethod
    def validate_alpha_beta(
        cls, v: Union[NDArray[Shape["1-*"], np.float64], None], info: ValidationInfo
    ) -> Union[NDArray[Shape["1-*"], np.float64], None]:
        """Validate the length of the alpha-beta kernel data."""
        if v is None:
            return v

        # Check if kernel is transposed but oterhwise of correct shape
        if (
            v.shape[0] == info.data["depths"].shape[0]
            and v.shape[1] == info.data["alpha_x"].size
            and v.shape[1] == info.data["beta_x"].size
        ):
            v = np.ascontiguousarray(v.T)

        if v.shape[0] != info.data["alpha_x"].size and v.shape[0] != info.data["beta_x"].size:
            raise ValueError(
                "Kernel data length does not match the number of reference photon alpha / beta "
                "values."
            )

        if v.shape[1] != info.data["depths"].size:
            raise ValueError("Kernel data length does not match the depth data length.")

        return v

    @computed_field(return_type=NDArray[Shape["1-*"], np.float64], alias="alphaBetaRatio")
    @property
    def alpha_beta_ratio(self) -> Union[NDArray[Shape["1-*"], np.float64], None]:
        """Photon alpha-beta ratio for each alpha / beta kernel."""
        if self.alpha is None or self.beta is None:
            return None
        return self.alpha_x / self.beta_x


class IonAccelerator(ExternalBeamMachine):
    """
    Machine Model for Ion Accelerators.

    Defines minimum meta-data an ion machine must hold
    Provides multiple data storage formats, currently supported
    - Pencil-Beam Kernels
    - Monte Carlo Beam Models.

    Attributes
    ----------
    sad : float
        The source-to-axis (-isocenter) distance of the machine
    """

    radiation_mode: Annotated[str, StringConstraints(pattern="^(protons|helium|carbon)$")] = Field(
        default="protons", validate_default=True
    )

    sad: float = Field(ge=0.0, description="Source-to-axis distance", alias="SAD")
    bams_to_iso_dist: float = Field(
        ge=0.0,
        description="Beam-monioring-system/nozzle to iso center distance",
        alias="BAMStoIsoDist",
    )
    lut_spot_size: NDArray[Shape["2,*"], np.float64] = Field(
        description="2D array of look up table latteral spotspacing to foxus index",
        alias=AliasChoices(
            "LUTspotSize", "LUT_bxWidthminFWHM"
        ),  # We have two aliases here to capture version changes in matRad base data files
    )
    # kernel_data: Annotated[np.recarray, Field(description="Pencil beam kernel data")]

    fit_air_offset: float = Field(
        ge=0.0,
        default=0.0,
        description="Air distance included in fitting of the kernel",
    )

    # Minimum required data is energies, Bragg-peak positions and beam foci
    peak_positions: NDArray[Shape["1-*"], np.float64]
    foci: dict[float, list[IonBeamFocus]]
    spectra: Optional[dict[float, IonEnergySpectrum]] = None

    # Optionally we can have pencil-beam kernels
    pb_kernels: Optional[dict[float, IonPencilBeamKernel]] = None

    @field_validator("lut_spot_size", mode="before")
    @classmethod
    def validate_possible_cast_array(cls, v: Any) -> Any:
        """Validate if the input can be cast to a float64 array."""
        try:
            v = np.array(v, dtype=np.float64)
        except ValueError as exc:
            raise exc

        return v

    # TODO: Beam model descriptions for Monte Carlo

    @classmethod
    def _parse_tabulated_energy_data_from_mat(
        cls, tabulated_energy_data: dict, returned_data: dict
    ):
        """Parse the tabulated energy data from a matRad machine file."""

        super(IonAccelerator, cls)._parse_tabulated_energy_data_from_mat(
            tabulated_energy_data, returned_data
        )

        # extrac required quantities
        if "offset" in tabulated_energy_data:
            returned_data["peak_positions"] = np.array(
                tabulated_energy_data["peakPos"], dtype=np.float64
            ) + np.array(tabulated_energy_data["offset"], dtype=np.float64)

        # extract beam foci
        foci = {}
        # First, check if we have the special situation of only a single focus

        foci_list = tabulated_energy_data["initFocus"]
        if not isinstance(foci_list, list):
            foci_list = [foci_list]

        for i, foci_entry in enumerate(foci_list):
            # check if all initFocus fields are lists
            if not all(isinstance(foci_entry[key], list) for key in foci_entry):
                focus_list = [foci_entry]
            else:
                focus_list = dl2ld(foci_entry)
            foci[returned_data["energies"][i]] = [IonBeamFocus(**focus) for focus in focus_list]

        returned_data["foci"] = foci

        # Energy Spectra
        spectra_list = tabulated_energy_data.get("energySpectrum", None)
        if spectra_list is not None:
            returned_data["spectra"] = {
                returned_data["energies"][i]: IonEnergySpectrumGaussian(**spectrum)
                for i, spectrum in enumerate(spectra_list)
            }

        # Extract the pencil beam kernels
        # Pencil beam kernels are not requried so we may skip them
        # if they can't be validated
        try:
            tmp = dl2ld(tabulated_energy_data)
            returned_data["pb_kernels"] = {
                entry["energy"]: IonPencilBeamKernel(**entry) for entry in tmp
            }
        except ValidationError:
            returned_data["pb_kernels"] = None

    # @model_validator(mode="after")
    # def validate_machine_components(self):
    #    pass
    # if self.betas not None:

    @model_validator(mode="after")
    def _check_machine(self) -> Self:
        """Validate the machine model for consistency."""

        if self.bams_to_iso_dist > self.sad:
            raise ValueError("BAMS to iso distance must be small than SAD.")

        if self.pb_kernels is not None:
            # Check that the peak_positions are consistent between peak_positions of the model
            # and the kernel peak positions with offset applied
            for i, energy in enumerate(self.energies):
                kernel = self.pb_kernels[energy]
                if not np.isclose(self.peak_positions[i], kernel.peak_pos + kernel.offset):
                    raise ValueError(
                        f"Peak position of the model and the kernel data for energy {energy} "
                        "are inconsistent."
                    )

        return self

    @property
    def has_pb_kernels(self):
        """Test for existance of valid pencil-beam kernels."""
        return self.pb_kernels is not None

    @property
    def pb_kernels_have_let(self) -> bool:
        """Check if the kernel data contains LET values."""
        return all(kernel.let is not None for kernel in self.pb_kernels.values())

    def get_kernel_by_index(self, ix_energy: int) -> IonPencilBeamKernel:
        """Get the pencil beam kernel for a specific energy index.

        Parameters
        ----------
        ix_energy : int
            Index of the energy level in the stored data

        Returns
        -------
        IonPencilBeamKernel
            Pencil beam kernel data
        """
        return self.get_kernel_by_energy(self.energies[ix_energy])

    def get_foci_by_index(self, ix_energy: int) -> list[IonBeamFocus]:
        """
        Get the focus data of the kernel for a specific energy index.

        Parameters
        ----------
        ix_energy : int
            Index of the energy level in the stored data

        Returns
        -------
        IonBeamFocus
            Dictionary of available beam foci
        """
        return self.foci[self.energies[ix_energy]]

    def get_kernel_by_energy(self, energy: float) -> IonPencilBeamKernel:
        """Get the pencil beam kernel for a specific energy value.

        Parameters
        ----------
        energy : float
            Energy value to search for

        Returns
        -------
        IonPencilBeamKernel
            Pencil beam kernel data
        """
        if self.pb_kernels is None:
            raise ValueError("No pencil beam kernels available for this machine.")
            # return None
        return self.pb_kernels[energy]

    def update_kernel_at_index(self, ix_energy: int, kernel: IonPencilBeamKernel):
        """Update the pencil beam kernel for a specific energy index.

        Parameters
        ----------
        ix_energy : int
            Index of the energy level in the stored data
        kernel : IonPencilBeamKernel
            Pencil beam kernel data
        """

        self.update_kernel_for_energy(self.energies[ix_energy], kernel)

    def update_kernel_for_energy(
        self, energy: Union[float, np.float64], kernel: IonPencilBeamKernel
    ):
        """
        Update a pencil-beam kernel for a specific energy.

        Parameters
        ----------
        energy : float
            Energy value to update the kernel for
        kernel : IonPencilBeamKernel
            The Kernel to replace the existing one

        Notes
        -----
        If your intention is to build a new machine / add kernels to a machine without exisiting
        kernels, you should create the machine from scratch or set the whole model field at once
        which is a dictionary of all kernels with the energy as key
        """
        if self.pb_kernels is None:
            raise ValueError("No pencil beam kernels available for this machine.")
        energy = np.float64(energy)

        self.pb_kernels[energy] = kernel

    def get_focus(self, ix_energy: int) -> dict:
        """
        Get the focus data of the kernel for a specific energy index.

        Parameters
        ----------
        ix_energy : int
            Index of the energy level in the stored data

        Returns
        -------
        dict
            Dictionary of focus data
        """

        return self.foci[self.energies[ix_energy]]

    @property
    def has_single_gaussian_kernel(self) -> bool:
        """
        Check if it has single gaussian pencil-beamkernels.

        Returns
        -------
        bool
            True if all kernels have single gaussian lateral scattering model
        """
        if self.pb_kernels is None:
            return False

        return all(kernel.sigma is not None for kernel in self.pb_kernels.values())

    @property
    def has_double_gaussian_kernel(self) -> bool:
        """
        Check if it has double gaussian pencil-beamkernels.

        Returns
        -------
        bool
            True if all kernels have double gaussian lateral scattering model
        """
        if self.pb_kernels is None:
            return False

        return all(
            kernel.sigma_1 is not None and kernel.sigma_2 is not None and kernel.weight is not None
            for kernel in self.pb_kernels.values()
        )

    @property
    def has_multi_gaussian_kernel(self) -> bool:
        """
        Check if it has multi gaussian pencil-beamkernels.

        Returns
        -------
        bool
            True if all kernels have multi gaussian lateral scattering model
        """
        if self.pb_kernels is None:
            return False

        return all(
            kernel.sigma_multi is not None and kernel.weight_multi is not None
            for kernel in self.pb_kernels.values()
        )

    @property
    def has_let_kernel(self) -> bool:
        """
        Check if it has LET values in the pencil-beam kernels.

        Returns
        -------
        bool
            True if all kernels have LET values
        """
        if self.pb_kernels is None:
            return False

        return all(kernel.let is not None for kernel in self.pb_kernels.values())

    @property
    def has_alpha_beta_kernels(self) -> bool:
        """
        Check if it has alpha-beta kernels.

        Returns
        -------
        bool
            True if all kernels have alpha-beta values
        """

        if self.pb_kernels is None:
            return False

        return all(
            kernel.alpha is not None and kernel.beta is not None
            for kernel in self.pb_kernels.values()
        )



"""Factory methods to manage available Problems."""

import warnings
import logging
from typing import Union, Type, Optional

from pyRadPlan.plan import Plan, validate_pln
from ._optiprob import PlanningProblem

PROBLEMS: dict[str, Type[PlanningProblem]] = {}

logger = logging.getLogger(__name__)


def register_problem(prob_cls: Type[PlanningProblem]) -> None:
    """
    Register a new problem.

    Parameters
    ----------
    prob_cls : type
        A PlanningProblem class.
    """
    if not issubclass(prob_cls, PlanningProblem):
        raise ValueError("PlanningProblem must be a subclass of PlanningProblem.")

    if prob_cls.short_name is None:
        raise ValueError("PlanningProblem must have a 'short_name' attribute.")

    if prob_cls.name is None:
        raise ValueError("PlanningProblem must have a 'name' attribute.")

    prob_name = prob_cls.short_name
    if prob_name in PROBLEMS:
        warnings.warn(f"PlanningProblem '{prob_name}' is already registered.")
    else:
        PROBLEMS[prob_name] = prob_cls


def get_available_problems(
    pln: Optional[Union[Plan, str]] = None,
) -> dict[str, Type[PlanningProblem]]:
    """
    Get a list of available planning problems based on the plan.

    Parameters
    ----------
    pln : Plan, optional
        A Plan object.

    Returns
    -------
    list
        A list of available planning problems.
    """

    if pln is None:
        return PROBLEMS
    pln = validate_pln(pln)
    return {
        name: cls
        for name, cls in PROBLEMS.items()
        if pln.radiation_mode in cls.possible_radiation_modes
    }


def get_problem(problem_desc: Union[str, dict, PlanningProblem]) -> Type[PlanningProblem]:
    """
    Returns a problem instance based on a descriptive parameter.

    Parameters
    ----------
    problem_desc : Union[str, dict, PlanningProblem, Plan]
        A string with the problem name, a dictionary with the problem configuration or a problem
        instance

    Returns
    -------
    PlanningProblem
        A problem instance
    """
    if isinstance(problem_desc, str):
        problem = PROBLEMS[problem_desc]()
    elif isinstance(problem_desc, dict):
        raise NotImplementedError(
            "PlanningProblem configuration from dictionary not implemented yet."
        )
    elif isinstance(problem_desc, PlanningProblem):
        problem = problem_desc
    else:
        raise ValueError(f"Invalid problem description: {problem_desc}")

    return problem


def get_problem_from_pln(pln: Union[Plan, dict]) -> Type[PlanningProblem]:
    """
    Factory function to get the appropriate problem based on the plan.

    Parameters
    ----------
    pln : Plan
        A Plan object.

    Returns
    -------
    Problem
        A Planning Problem object.
    """
    pln = validate_pln(pln)
    problems = get_available_problems(pln)

    if len(problems) <= 0:
        raise ValueError(
            f"No PlanningProblem available for radiation mode '{pln.radiation_mode}'."
        )

    problem_names = list(problems.keys())

    # Did the user provide an problem in the pln?
    if isinstance(pln.prop_opt, PlanningProblem):
        # The user provided an problem object, so lets use it
        # but warn the user if it is not in the available problems
        problem_name = pln.prop_opt.short_name
        if problem_name not in problems:
            warnings.warn(f"Engine '{problem_name}' seems not to be valid for Plan setup.")
        return pln.prop_opt

    if isinstance(pln.prop_opt, dict):
        # The user provided a dictionary with problem parameters,
        # so we need to find the problem name
        if "problem" in pln.prop_opt:
            if pln.prop_opt["problem"] in problems:
                return problems[pln.prop_opt["problem"]](pln)
            warnings.warn(f"Engine '{pln.prop_opt['problem']}' not available for Plan.")

        # If no problem name was found, we choose the first as default
        logger.warning(
            "No PlanningProblem specified in Plan. Using first available problem %s.",
            problem_names[0],
        )
        return problems[problem_names[0]](pln)

    raise ValueError(f"No problem available for radiation mode '{pln.radiation_mode}'.")



"""Test Helper functions for pyRadPlan."""

from typing import Union, get_type_hints, get_origin, get_args
import warnings
import numpy as np
from pydantic import BaseModel
from pydantic.fields import FieldInfo, ComputedFieldInfo
from pyRadPlan.core import PyRadPlanBaseModel
import scipy.sparse as sp


def dl2ld(dict_of_lists: dict[str, list], type_check: bool = True) -> list[dict]:
    """Converts a dictionary of lists to a list of dictionaries.

    Parameters
    ----------
    dict_of_lists : dict
        The dictionary of lists to convert.
    type_check : bool, optional
        Whether to perform type checking, by default True

    Returns
    -------
    list[dict]
        A list of dictionaries.
    """

    if type_check:
        # Check if dict_of_lists is a dictionary
        if not isinstance(dict_of_lists, dict):
            raise TypeError("The input must be a dictionary.")
        # Check if all values in the dictionary are lists
        if not all(isinstance(value, list) for value in dict_of_lists.values()):
            raise TypeError("All values in the dictionary must be lists.")
        # CHeck if all lists have the same length
        if len(set(map(len, dict_of_lists.values()))) > 1:
            raise TypeError("All lists in the dictionary must have the same length.")

    # Empty input returns empty array
    if dict_of_lists == {}:
        return []

    # Convert the dictionary of lists to a list of dictionaries
    list_of_dicts = [dict(zip(dict_of_lists, t)) for t in zip(*dict_of_lists.values())]

    return list_of_dicts


def ld2dl(list_of_dicts: list[dict], type_check: bool = True) -> dict[str, list]:
    """Converts a list of dictionaries to a dictionary of lists.

    Parameters
    ----------
    list_of_dicts : list
        The list of dictionaries to convert.
    type_check : bool, optional
        Whether to perform type checking, by default True

    Returns
    -------
    dict
        A dictionary of lists.
    """

    if type_check:
        # Check if list_of_dicts is a list
        if not isinstance(list_of_dicts, list):
            raise TypeError("The input must be a list.")
        # Check if all elements in the list are dictionaries
        if not all(isinstance(element, dict) for element in list_of_dicts):
            raise TypeError("All elements in the list must be dictionaries.")

    # Empty input returns empty array
    if len(list_of_dicts) == 0:
        return {}

    # Convert the list of dictionaries to a dictionary of lists
    dict_of_lists = {key: [element[key] for element in list_of_dicts] for key in list_of_dicts[0]}

    return dict_of_lists


def models2recarray(
    models: list[BaseModel],
    serialization_context: Union[dict, str] = None,
    override_types: dict = None,
    by_alias: bool = False,
) -> np.recarray:
    """Converts a list of PyRadPlanBaseModel instances to a numpy recarray.

    Parameters
    ----------
    models : list
        The list of PyRadPlanBaseModel instances to convert.
    serialization_context : str, optional
        The context in which the datastructure should be serialized, by default None.
    override_types : dict, optional
        A dictionary of types to override the automatically obtained types, by default None.
        If the context contains "matRad" and the models are of type PyRadPlanBaseModel, the models
        will be converted to matRad compatible structures.
    by_alias : bool, optional
        Whether to use the alias names for serialization, by default False.

    Returns
    -------
    np.recarray
        A numpy recarray.
    """

    # Check if models is a list
    if not isinstance(models, list):
        raise TypeError("The input must be a list.")

    # Check if models is not empty
    if not models:
        raise ValueError("The input list must not be empty.")

    first_model_type = type(models[0])

    # Check if the type is a subclass of a BaseModel
    if not issubclass(first_model_type, BaseModel):
        raise TypeError("All elements in the list must be subclasses of BaseModel.")

    # Check if all modelss in the list are of the same type
    if not all(isinstance(model, first_model_type) for model in models):
        raise TypeError("All models in the list must be of the same type.")

    if "matRad" in serialization_context and isinstance(models[0], PyRadPlanBaseModel):
        models_dump = [model.to_matrad(context=serialization_context) for model in models]
        by_alias = True
    else:
        models_dump = [
            model.model_dump(by_alias=by_alias, context=serialization_context) for model in models
        ]

    model_fields = models[0].model_fields | models[0].model_computed_fields
    model_field_types = get_type_hints(models[0].__class__)
    for cfield in models[0].model_computed_fields:
        model_field_types.update({cfield: model_fields[cfield].return_type})

    # Debug
    if "machine" in model_fields:
        pass

    # Type management
    # Override manual types
    if override_types:
        model_field_types.update(override_types)

        # Remove model_field and model_field_types where the override_types is None
        model_fields = {
            field: model_fields[field]
            for field in model_fields
            if model_field_types[field] is not None
        }
        model_field_types = {
            field: model_field_types[field]
            for field in model_field_types
            if model_field_types[field] is not None
        }

    # convert dtypes
    for field in model_field_types:
        # str -> 'U'
        if model_field_types[field] is str:
            model_field_types[field] = object

        # list management
        if get_origin(model_field_types[field]) is list:
            model_field_types[field] = np.ndarray

        # manage optional types
        if get_origin(model_field_types[field]) is Union:
            t_args = get_args(model_field_types[field])
            if type(None) in t_args and len(t_args) == 2:
                model_field_types[field] = t_args[0]
            else:
                warnings.warn(
                    f"Field {field} has a Union type that is not supported. Will use object."
                )
                model_field_types[field] = object

    aliases = {}
    for field in model_fields:
        if by_alias:
            if isinstance(model_fields[field], FieldInfo):
                aliases[field] = model_fields[field].serialization_alias
            elif isinstance(model_fields[field], ComputedFieldInfo):
                aliases[field] = model_fields[field].alias
            else:  # Sanity Check
                raise TypeError(f"Field {field} is not a FieldInfo or ComputedFieldInfo object.")
        else:
            aliases[field] = field

    # Create a dtype for the structured armodel
    models_dtype = np.dtype([(aliases[field], model_field_types[field]) for field in model_fields])

    models_recarray = np.recarray((len(models),), dtype=models_dtype)

    for i, model_dict in enumerate(models_dump):
        for field in model_fields:
            fname = aliases[field]
            value = model_dict[fname]

            if model_field_types[field] == np.str_:
                value = np.str_(value)
            elif model_field_types[field] == np.ndarray:
                value = np.asarray(value)
            models_recarray[fname][i] = value

    return models_recarray


def swap_orientation_sparse_matrix(
    sparse_matrix: sp.csc_matrix, original_shape, axes
) -> sp.csc_matrix:
    """
    Swaps the specified axes of a sparse matrix.

    Parameters
    ----------
    sparse_matrix : sp.csc_matrix
        The sparse matrix to swap axes.
    original_shape : tuple
        The original shape of the matrix.
    axes : tuple
        The axes to swap.

    Returns
    -------
    sp.csc_matrix
        The sparse matrix with swapped axes.
    """
    row_indices, _ = sparse_matrix.nonzero()

    if axes in ((0, 1), (1, 0)):
        j, i, k = np.unravel_index(row_indices, original_shape)
        new_shape = (original_shape[1], original_shape[0], original_shape[2])
    elif axes in ((0, 2), (2, 0)):
        k, j, i = np.unravel_index(row_indices, original_shape)
        new_shape = (original_shape[2], original_shape[1], original_shape[0])
    elif axes in ((1, 2), (2, 1)):
        i, k, j = np.unravel_index(row_indices, original_shape)
        new_shape = (original_shape[0], original_shape[2], original_shape[1])
    else:
        raise ValueError("Invalid axes for swapping")

    new_indices = np.ravel_multi_index((i, j, k), new_shape)
    num_rows = np.prod(original_shape)
    permutation = sp.csc_matrix(
        (np.ones_like(row_indices), (new_indices, row_indices)),
        shape=(num_rows, num_rows),
        dtype=bool,
    )

    reordered_sparse_matrix = permutation @ sparse_matrix
    return reordered_sparse_matrix



import logging

import numpy as np

from pyRadPlan import (
    PhotonPlan,
    load_tg119,
    generate_stf,
    calc_dose_influence,
    fluence_optimization,
    plot_slice,
)

from pyRadPlan.optimization.objectives import SquaredDeviation, SquaredOverdosing, MeanDose

logging.basicConfig(level=logging.INFO)

# Load TG119 (provided within pyRadPlan)
ct, cst = load_tg119()

# Create a plan object
pln = PhotonPlan(machine="Generic")
num_of_beams = 5
pln.prop_stf = {
    "gantry_angles": np.linspace(0, 360, num_of_beams, endpoint=False),
    "couch_angles": np.zeros((num_of_beams,)),
}
# Generate Steering Geometry ("stf")
stf = generate_stf(ct, cst, pln)

# Calculate Dose Influence Matrix ("dij")
dij = calc_dose_influence(ct, cst, stf, pln)

# Optimize
cst.vois[1].objectives = [SquaredDeviation(priority=1000.0, d_ref=3.0)]  # Target
cst.vois[0].objectives = [SquaredOverdosing(priority=100.0, d_max=1.0)]  # OAR
cst.vois[2].objectives = [
    MeanDose(priority=1.0, d_ref=0.0),
    SquaredOverdosing(priority=10.0, d_max=2.0),
]  # BODY

fluence = fluence_optimization(ct, cst, stf, dij, pln)

# Result
result = dij.compute_result_ct_grid(fluence)

# Choose a slice to visualize
view_slice = int(np.round(ct.size[2] / 2))

# Visualize
plot_slice(
    ct=ct,
    cst=cst,
    overlay=result["physical_dose"],
    view_slice=view_slice,
    plane="axial",
    overlay_unit="Gy",
)



"""Core module with fundamental classes and functions for pyRadPlan."""

from ._exceptions import PyRadPlanError
from .datamodel import PyRadPlanBaseModel
from .resample import resample_image, resample_numpy_array
from ._grids import Grid

__all__ = [
    "PyRadPlanError",
    "np2sitk",
    "resample_image",
    "resample_numpy_array",
    "PyRadPlanBaseModel",
    "Grid",
]



from pyRadPlan.stf import IonSpot, PhotonBixel
from pyRadPlan.stf import RangeShifter


def test_proton_beamlet_only_energy():
    proton_beamlet = IonSpot(energy=100)
    assert proton_beamlet.energy == 100
    assert proton_beamlet.num_particles_per_mu == 1.0e6
    assert proton_beamlet.min_mu == 0.0
    assert proton_beamlet.max_mu == float("inf")
    assert isinstance(proton_beamlet.range_shifter, RangeShifter)
    assert proton_beamlet.focus_ix == 0


def test_proton_beamlet_all():
    proton_beamlet = IonSpot(
        energy=100,
        num_particles_per_mu=1.0e7,
        min_mu=1.0,
        max_mu=10.0,
        range_shifter=RangeShifter(),
        focus_ix=1,
    )
    assert proton_beamlet.energy == 100
    assert proton_beamlet.num_particles_per_mu == 1.0e7
    assert proton_beamlet.min_mu == 1.0
    assert proton_beamlet.max_mu == 10.0
    assert isinstance(proton_beamlet.range_shifter, RangeShifter)
    assert proton_beamlet.focus_ix == 1


def test_photon_beamlet_only_energy():
    photon_beamlet = PhotonBixel(energy=6)
    assert photon_beamlet.energy == 6
    assert photon_beamlet.num_particles_per_mu == 1.0e6
    assert photon_beamlet.min_mu == 0.0
    assert photon_beamlet.max_mu == float("inf")
    assert photon_beamlet.relative_fluence == 1.0


def test_photon_beamlet_all():
    photon_beamlet = PhotonBixel(
        energy=6,
        num_particles_per_mu=1.0e7,
        min_mu=1.0,
        max_mu=10.0,
        relative_fluence=0.5,
    )
    assert photon_beamlet.energy == 6
    assert photon_beamlet.num_particles_per_mu == 1.0e7
    assert photon_beamlet.min_mu == 1.0
    assert photon_beamlet.max_mu == 10.0
    assert photon_beamlet.relative_fluence == 0.5



from numpy.typing import NDArray
import pint

from pyRadPlan.quantities._base import FluenceDependentQuantity

ureg = pint.UnitRegistry()


class Dose(FluenceDependentQuantity):
    """Dose quantity depending on fluence."""

    unit = ureg.gray
    dim = 1
    identifier = "physical_dose"
    name = "dose"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.identifier not in self._dij.model_fields:
            raise ValueError(f"Quantity {self.identifier} not available in Dij object.")

        # TODO: We could also check if it is an array here

    def _compute_quantity_single_scenario_from_cache(self, scenario_index: int) -> NDArray:
        return self._dij.physical_dose.flat[scenario_index] @ self._w_cache

    def _compute_chained_fluence_gradient_single_scenario_from_cache(
        self, d_quantity, scenario_index: int
    ) -> NDArray:
        return d_quantity @ self._dij.physical_dose.flat[scenario_index]



from pyRadPlan.optimization.problems import (
    NonLinearFluencePlanningProblem,
    NonLinearPlanningProblem,
    PlanningProblem,
)
from pyRadPlan.plan import create_pln


def test_construct():
    pln = create_pln({"radiation_mode": "protons", "machine": "Generic"})
    prob = NonLinearFluencePlanningProblem(pln)

    assert isinstance(prob, NonLinearFluencePlanningProblem)
    assert isinstance(prob, NonLinearPlanningProblem)
    assert isinstance(prob, PlanningProblem)


def test_construct_noplan():
    prob = NonLinearFluencePlanningProblem()

    assert isinstance(prob, NonLinearFluencePlanningProblem)
    assert isinstance(prob, NonLinearPlanningProblem)
    assert isinstance(prob, PlanningProblem)



import pytest
from pyRadPlan.util import dl2ld, ld2dl, swap_orientation_sparse_matrix
import scipy.sparse as sp
import numpy as np


### Test dl2ld function
def test_dl2ld_valid_input():
    dict_of_lists = {"a": [1, 2, 3], "b": [4, 5, 6]}
    expected_output = [{"a": 1, "b": 4}, {"a": 2, "b": 5}, {"a": 3, "b": 6}]
    assert dl2ld(dict_of_lists) == expected_output


def test_dl2ld_empty_input():
    dict_of_lists = {}
    assert dl2ld(dict_of_lists) == []


def test_dl2ld_type_check_false():
    dict_of_lists = {"a": [1, 2, 3], "b": [4, 5, 6]}
    expected_output = [{"a": 1, "b": 4}, {"a": 2, "b": 5}, {"a": 3, "b": 6}]
    assert dl2ld(dict_of_lists, type_check=False) == expected_output


def test_dl2ld_type_error_not_dict():
    with pytest.raises(TypeError):
        dl2ld([1, 2, 3])


def test_dl2ld_type_error_values_not_lists():
    dict_of_lists = {"a": [1, 2, 3], "b": 4}
    with pytest.raises(TypeError):
        dl2ld(dict_of_lists)


def test_dl2ld_type_error_lists_different_lengths():
    dict_of_lists = {"a": [1, 2, 3], "b": [4, 5]}
    with pytest.raises(TypeError):
        dl2ld(dict_of_lists)


### Test ld2dl function
def test_ld2dl_valid_input():
    list_of_dicts = [{"a": 1, "b": 4}, {"a": 2, "b": 5}, {"a": 3, "b": 6}]
    expected_output = {"a": [1, 2, 3], "b": [4, 5, 6]}
    assert ld2dl(list_of_dicts) == expected_output


def test_ld2dl_empty_input():
    list_of_dicts = []
    assert ld2dl(list_of_dicts) == {}


def test_ld2dl_type_check_false():
    list_of_dicts = [{"a": 1, "b": 4}, {"a": 2, "b": 5}, {"a": 3, "b": 6}]
    expected_output = {"a": [1, 2, 3], "b": [4, 5, 6]}
    assert ld2dl(list_of_dicts, type_check=False) == expected_output


def test_ld2dl_type_error_not_list():
    with pytest.raises(TypeError):
        ld2dl({"a": 1, "b": 4})


def test_ld2dl_type_error_elements_not_dicts():
    list_of_dicts = [{"a": 1, "b": 4}, [2, 5], {"a": 3, "b": 6}]
    with pytest.raises(TypeError):
        ld2dl(list_of_dicts)


def test_swap_orientation_sparse_matrix():
    original_shape = (2, 3, 4)
    data = [1, 2, 3]
    row_indices = [0, 1, 2]
    col_indices = [1, 2, 3]

    sparse_matrix = sp.csc_matrix(
        (data, (row_indices, col_indices)), shape=(np.prod(original_shape), original_shape[2])
    )
    axes = (0, 1)

    result = swap_orientation_sparse_matrix(sparse_matrix, original_shape, axes)

    result_reversed = swap_orientation_sparse_matrix(
        result, (original_shape[1], original_shape[0], original_shape[2]), axes
    )

    result_data = result.data

    assert np.array_equal(sparse_matrix.toarray(), result_reversed.toarray())
    assert sp.issparse(result)
    assert np.array_equal(data, result_data)
    assert result.shape == sparse_matrix.shape
